/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/26/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __fs_posix_api_h__
#define __fs_posix_api_h__ 1

/*   std includes   */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>

/*   project headers   */
#include "fs-status.h"
#include "fs-dir.h"


namespace duct {
  
  namespace filesystem {
    
    
    
    namespace __posix_api {
    
 
      /*   setting permissions   */
      inline void permissions(const path& p, perms prms, int& errval) noexcept {
        if(::chmod(p.c_str(), static_cast<::mode_t>(prms & perms::mask))) errval = errno;
      }
      
      /*   read symlink   */
      path read_symlink(const path&, int&) noexcept;
      
      /*   setting the last write time   */
      void last_write_time(const path&, const file_time_type, int&) noexcept;
 
      
      /*   creating a directory   */
      bool create_directory(const path&, int&) noexcept;
      bool create_directory(const path&, const path&, int&) noexcept;
      
      
      /*   creating a symlink and hardlink   */
      inline void create_symlink(const path& to, const path& from, int& errval) noexcept {
        if(::symlink(to.c_str(), from.c_str())) errval = errno;
      }

      inline void create_directory_symlink(const path& to, const path& from, int& errval) noexcept {
        if(::symlink(to.c_str(), from.c_str())) errval = errno;
      }
      
      inline void create_hard_link(const path& to, const path& from, int& errval) noexcept {
        if(::link(to.c_str(), from.c_str())) errval = errno;
      }
      
      
      /*    current path   */
      path current_path(int&) noexcept;

      inline void current_path(const path& p, int& errval) noexcept {
        if(::chdir(p.c_str())) errval = errno;
      }
      
      /*   remove   */
      bool remove(const path&, file_type, int&) noexcept;

      
      /*    copy a file  */
      void copy_file(const std::string&, const std::string&, int&) noexcept;
      
      
      /*  rename file  */
      inline void rename(const path& from, const path& to, int& errval) noexcept {
        if(::rename(from.c_str(), to.c_str())) errval = errno;
      }

      
      /*  implementing directory_iterator   */
      void dir_itr_close(directory_iterator::dir_itr *);
      directory_iterator::dir_itr * dir_itr_ctr(const path&, int&);
      void dir_itr_incr(std::shared_ptr<directory_iterator::dir_itr>&, int&);

    }
  }
}

#endif
