/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 04/11/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_fs_search_path_h__
#define __cxx_fs_search_path_h__ 1


/*   std includes   */
#include <list>

/*   project headers   */
#include <cxx/filesystem/fs-path.h>



namespace duct {
  namespace filesystem {
    
    
    class search_path
    {
    public:
      /*   constructors   */
      search_path() = default;
      
      explicit search_path(path&& p) { _M_list.push_back(std::move(p));}
      explicit search_path(const path& p) { _M_list.push_back(p);}
      
      template<class _Path1, class _Path2, class ..._Rest>
      search_path(_Path1&& p1, _Path2&& p2, _Rest&& ...r) {
        this->_M_add(std::forward<_Path1>(p1), std::forward<_Path2>(p2), std::forward<_Rest>(r)...);
      }
      
      
      /*   try to find a file or directory    */
      path find(const path& file) const
      {
        for(const auto& x : _M_list) {
          path f{x/file};
          if(exists(f)) return f;
        }
        
        return path{};
      }
      
      /*   add  elements to the search path   */
      void add_path(path&& p) { _M_list.push_back(std::move(p));}
      void add_path(const path& p) { _M_list.push_back(p);}
      
      /*   copy elements from another search_path object    */
      void add_path(const search_path& p) {
        for(const auto& x : p._M_list)
          _M_list.push_back(x);
      }
      
      /*   add elements recursively    */
      template<class _Path1, class _Path2, class ..._Rest>
      void add_path(_Path1&& p1, _Path2&& p2, _Rest&& ...r) {
        this->_M_add(std::forward<_Path1>(p1), std::forward<_Path2>(p2), std::forward<_Rest>(r)...);
      }
      
    private:
      /*   list of the PATH elements    */
      std::list<path> _M_list;
      
      /*   helper function to implement the add_path function   */
      template<class _Path, class ..._Rest>
      void _M_add(_Path&& p, _Rest&&... r) {
        this -> add_path(std::forward<_Path>(p));
        this -> _M_add(std::forward<_Rest>(r)...);
      }
      
      template<class _Path>
      void _M_add(_Path&& p) {
        this -> add_path(std::forward<_Path>(p));
      }
    };
    
  }
}

#endif


