/*  Copyright (C) 2010-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 10/5/10.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   standard c++ includes
#include <limits>

//   nlojet++ includes
#include "bits/duct-mapping.h"



namespace duct {
  
  //========> Some helper fuctions to compute the mappings <=====
  namespace __details {
    
    
#define QUARK particle_state::colorlink::quarkend
#define ANTI_QUARK particle_state::colorlink::antiquarkend
    
    //        Elementary color operators
    template<std::size_t _Nin>
    static void color_a_dagger(int sigma, int emitter, event_base<_Nin>& p)
    {
      int spectator, m = p.template number_of<qcd_out>();
      
      switch(sigma) {
        case 1:
          spectator = p[emitter].color.next;
          p[emitter].color.next  = m;
          p[m].color.prev = emitter;
          p[m].color.next = spectator;
          p[spectator].color.prev = m;
          break;
        case -1:
          spectator = p[emitter].color.prev;
          p[spectator].color.next = m;
          p[m].color.prev = spectator;
          p[m].color.next = emitter;
          p[emitter].color.prev = m;
          break;
        case 2:
          spectator = p[emitter].color.prev;
          p[emitter].color.prev  = QUARK;
          p[spectator].color.next  = m;
          p[m].color.prev = spectator;
          p[m].color.next = ANTI_QUARK;
          break;
        case -2:
          p[p[emitter].color.prev].color.next = p[emitter].color.next;
          p[p[emitter].color.next].color.prev = p[emitter].color.prev;
          p[emitter].color.prev = QUARK;
          p[emitter].color.next  = m;
          p[m].color.prev = emitter;
          p[m].color.next = ANTI_QUARK;
          break;
        case 3:
          spectator = p[emitter].color.next;
          p[emitter].color.next = ANTI_QUARK;
          p[spectator].color.prev = m;
          p[m].color.next = spectator;
          p[m].color.prev = QUARK;
          break;
        case -3:
          p[p[emitter].color.prev].color.next = p[emitter].color.next;
          p[p[emitter].color.next].color.prev = p[emitter].color.prev;
          p[emitter].color.next = ANTI_QUARK;
          p[emitter].color.prev  = m;
          p[m].color.next = emitter;
          p[m].color.prev = QUARK;
          break;
        case 4:
          p[emitter].color.prev  = m;
          p[m].color.next = emitter;
          p[m].color.prev = QUARK;
          break;
        case -4:
          p[emitter].color.next  = m;
          p[m].color.prev = emitter;
          p[m].color.next = ANTI_QUARK;
          break;
        default:
          //   do nothing
          break;
      }
    }
    
    template<std::size_t _Nin>
    static void color_a(int sigma, int emitter, event_base<_Nin>& p)
    {
      int spectator, m = p.template number_of<qcd_out>();
      
      switch(sigma) {
        case 1:
          spectator = p[emitter].colorP.next;
          p[emitter].colorP.next  = m;
          p[m].colorP.prev = emitter;
          p[m].colorP.next = spectator;
          p[spectator].colorP.prev = m;
          break;
        case -1:
          spectator = p[emitter].colorP.prev;
          p[spectator].colorP.next = m;
          p[m].colorP.prev = spectator;
          p[m].colorP.next = emitter;
          p[emitter].colorP.prev = m;
          break;
        case 2:
          spectator = p[emitter].colorP.prev;
          p[emitter].colorP.prev  = QUARK;
          p[spectator].colorP.next  = m;
          p[m].colorP.prev = spectator;
          p[m].colorP.next = ANTI_QUARK;
          break;
        case -2:
          p[p[emitter].colorP.prev].colorP.next = p[emitter].colorP.next;
          p[p[emitter].colorP.next].colorP.prev = p[emitter].colorP.prev;
          p[emitter].colorP.prev = QUARK;
          p[emitter].colorP.next  = m;
          p[m].colorP.prev = emitter;
          p[m].colorP.next = ANTI_QUARK;
          break;
        case 3:
          spectator = p[emitter].colorP.next;
          p[emitter].colorP.next = ANTI_QUARK;
          p[spectator].colorP.prev = m;
          p[m].colorP.next = spectator;
          p[m].colorP.prev = QUARK;
          break;
        case -3:
          p[p[emitter].colorP.prev].colorP.next = p[emitter].colorP.next;
          p[p[emitter].colorP.next].colorP.prev = p[emitter].colorP.prev;
          p[emitter].colorP.next = ANTI_QUARK;
          p[emitter].colorP.prev  = m;
          p[m].colorP.next = emitter;
          p[m].colorP.prev = QUARK;
          break;
        case 4:
          p[emitter].colorP.prev  = m;
          p[m].colorP.next = emitter;
          p[m].colorP.prev = QUARK;
          break;
        case -4:
          p[emitter].colorP.next  = m;
          p[m].colorP.prev = emitter;
          p[m].colorP.next = ANTI_QUARK;
          break;
        default:
          //  do nothing
          break;
      }
    }
    

    /*  Lorentz transformation, that takes care about the special case when the
     *  spectator sector is massles.
     */
    template<std::size_t _Nin, class _Boost>
    inline void boost_fin(_Boost&& boost, int emitter, event_base<_Nin>& p)
    {
      /*    boosting the non-qcd momenta    */
      for(particle_state& ki : p.template range<non_qcd>()) {
        ki.momentum = boost(ki.momentum);
      }
      
      /*    boosting the qcd momenta    */
      auto qcd = p.template range<qcd_out>();
      auto p1 = qcd.begin();
      auto pl = p1 + emitter-1, pn = qcd.end()-1;
      
      for(auto pi = p1; pi < pn; pi++)
        if(pi != pl) pi->momentum = boost(pi->momentum);
    }
    
  
    /*  Lorentz transformation in the initial state. One could use the standard
     *  transformation but this one is a little bit better for the initial state.
     */
    template<std::size_t _Nin>
    static void boost_ini(const lorentzvector<double>& na, const lorentzvector<double>& nb,
                          double omega, const lorentzvector<double>& vT, event_base<_Nin>& p)
    {
      double s = na*nb, vT2 = vT.mag2();
      double rs = std::sqrt(2.0*s);
      double alpha, beta, twors = 2.0/rs;
      lorentzvector<double> pT;
      
      
      /*      boosting the non-qcd momenta     */
      for(particle_state& ki : p.template range<non_qcd>()) {
        alpha = ki.momentum*nb/s;
        beta = ki.momentum*na/s;
        pT = ki.momentum - alpha*na - beta*nb;
        
        beta -= alpha*vT2 + twors*(vT*pT);
        beta /= omega;
        pT += alpha*rs*vT;
        alpha *= omega;
        
        ki.momentum = alpha*na + beta*nb + pT;
      }
      
      /*    boosting the qcd momenta    */
      auto qcd = p.template range<qcd_out>();
      auto p1 = qcd.begin(), pn = qcd.end()-1;
      
      for(auto pi = p1; pi < pn; pi++) {
        alpha = pi->momentum*nb/s;
        beta = pi->momentum*na/s;
        pT = pi->momentum - alpha*na - beta*nb;
        
        beta -= alpha*vT2 + twors*(vT*pT);
        beta /= omega;
        pT += alpha*rs*vT;
        alpha *= omega;
        
        pi->momentum = alpha*na + beta*nb + pT;
      }
    }
    
  
    /*  Momentum mapping in the final state shower.
     */
    template<std::size_t _Nin>
    static std::function<lorentzvector<double>(const lorentzvector<double>&)>
    momentum_mapping_fin(event_base<_Nin>& p, const event_base<_Nin>& q, const splitting_variables<shower<lo>>& vars)
    {
      //----- number of partons & particle table -----
      unsigned int np = q.template number_of<qcd_out>();
      unsigned int nnqcd = q.template number_of<non_qcd>();
      unsigned int ninc = q.template number_of<incomings>();
      model_ref mdl = q.get_model();
      
      //----- copying the spectator momenta -----
      {
        //----- resize if it is needed ---
        p.resize(nnqcd, np+1);
        
        //----- matching the particle tables -----
        p.set_model(mdl);
        p.color_suppression_index = q.color_suppression_index; 
      
        //----- copying the state q into p ---- 
        typename event_base<_Nin>::iterator pi = p.begin();
        typename event_base<_Nin>::const_iterator qi = q.begin();
        
        for(; qi < q.end(); ++pi, ++qi) *pi = *qi;
      }
      
      //----- iterators -----
      typename event_base<_Nin>::iterator pb = p.begin();
      typename event_base<_Nin>::iterator p1 = pb+nnqcd+ninc;
      typename event_base<_Nin>::iterator pn = p1+np, pl = p1+vars.emitter-1;
      
      /*    total incoming momentum &  the mass of the daughters   */
      lorentzvector<double> Q = p.template momentum_sum<incomings>();
      double pQ = 2.0*(pl->momentum*Q), Q2 = Q.mag2();
      
      double b1 = (6 == std::abs(vars.fl) ? mdl.mass2(vars.fl)/pQ : 0.0);
      double b2 = (6 == std::abs(vars.fm) ? mdl.mass2(vars.fm)/pQ : 0.0);
 
      /*   basic hard parameters: al, bl and rl    */
      double al = Q2/pQ, bl = pl->momentum.mag2()/pQ;
      double rl = std::sqrt(std::abs(1.0-4.0*al*bl)), tmp = (1.0-bl)/(std::sqrt(al)+std::sqrt(std::abs(al+bl-1.0)));
      double ymax = tmp*tmp-bl;
      
      
      /*    y dependent variables    */
      double y = vars.y;
      double lmd = std::sqrt(std::abs((1.0-y/ymax)*(rl*rl-ymax*y)));
      double x0 = (1.0+y-lmd)/(1.0+y+lmd), Fg = (1.0+y + lmd)*(1.0+rl)/(4.0*al*rl);
      double zp = 1.0, zm  = 0.0;
      
      if(b2 == 0.0)
        zm = b1/(y+bl);
      else if(b1 == b2) {
        zp = 0.5 + 0.5*std::sqrt(std::abs(1.0-4.0*b1/(y+bl)));
        zm = 1.0-zp;
      } else {
        double yl12 = y+bl+b1-b2, yl = y+bl;
        double sr = std::sqrt(std::abs(yl12*yl12-4.0*yl*b1));
        zp = 0.5*(yl12+sr)/yl;
        zm = 0.5*(yl12-sr)/yl;
      }	  
      
      /*    momenta kl, nl and Pl    */
      lorentzvector<double> nl = Q - (2.0*al/(1.0+rl))*pl->momentum;
      lorentzvector<double> kl = 2.0*al/(1.0+rl)*pl->momentum - (1.0-rl)/(1.0+rl)*Q;
      lorentzvector<double> Pl = Fg*(kl + x0*nl);
      
  
      /*    boost the spectators    */
      lorentzvector<double> hK = Q-Pl, K = Q-pl->momentum;
      double op, om, nk = kl*nl;
      
      if(2 == np + nnqcd && K.mag2() < 1e-9*Q2) {
        double ratio = (K*kl)/(hK*kl);
        
        op = ratio-1.0;
        om = 1.0/ratio-1.0;
      }
      else {
        op = (rl-y-lmd)/(2.0*al-1.0-rl);
        om = (lmd-rl-y)/(2.0*al-1.0+rl);
      }
      
      auto boost = [=](const lorentzvector<double>& qnull) -> lorentzvector<double> {
        return qnull + kl*(op*(nl*qnull)/nk) + nl*(om*(kl*qnull)/nk);
      };
      
      __details::boost_fin(boost, vars.emitter, p);

      
      /*    the transverse momentum in the c.m. of Pl   */
      double z = (zp-zm)*vars.z + zm;
      double mT = std::sqrt(std::abs((z-zm)*(zp-z)*x0*2.0*nk));
      lorentzvector<double> kT(mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0, 0.0);
        
      /*    boosting the transverse momentum back to the original frame    */
      lorentzvector<double> ql = kl, qk = p[vars.spectator].momentum;
      threevector<double> bvec = Pl.boostVector();
      
      ql.boost(-bvec);
      double rPhi = ql.phi(), rTheta = ql.theta();
      
      if(vars.spectator != vars.emitter) {
        qk.boost(-bvec);
        qk.rotateZ(-rPhi);
        qk.rotateY(-rTheta);
        kT.rotateZ(qk.phi());
      }
      
      kT.rotateY(rTheta);
      kT.rotateZ(rPhi);
      kT.boost(bvec);
      
      /*    the daughter momenta & flavors    */
      pl->flavor = vars.fl;
      pl->momentum = Fg*(z*kl + (zp+zm-z)*x0*nl - kT);
      
      pn->flavor = vars.fm;
      pn->momentum = Pl-pl->momentum;
      
      return boost;
    }
    
    
    /*  Momentum mapping in the initial state shower.
     */
    template<std::size_t _Nin>
    static std::function<lorentzvector<double>(const lorentzvector<double>&)>
    momentum_mapping_ini(event_base<_Nin>& p, const lorentzvector<double>& pA, const lorentzvector<double>& pB,
                         const event_base<_Nin>& q, const splitting_variables<shower<lo>>& vars)
    {
      //----- number of partons & particle table -----
      unsigned int np = q.template number_of<qcd_out>();
      unsigned int nnqcd = q.template number_of<non_qcd>();
      model_ref mdl = q.get_model();
      
      //----- copying the spectator momenta -----
      {
        /*     resize the new event   */
        p.resize(nnqcd, np+1);
        
        /*    matching the particle tables   */
        p.set_model(mdl);
        p.color_suppression_index = q.color_suppression_index; 
              
        //----- copying the state q into p ----
        typename event_base<_Nin>::iterator pi = p.begin();
        typename event_base<_Nin>::const_iterator qi = q.begin();
        
        for(; qi < q.end(); ++pi, ++qi) *pi = *qi;
      }
      
      //----- iterators -----
      typename event_base<_Nin>::iterator pa = p.begin()+nnqcd;
      typename event_base<_Nin>::iterator pb = pa+1, pn = p.end()-1;
      
      //----- incoming momentum ----
      if(vars.emitter == 0) std::swap(pa,pb);
      lorentzvector<double> na = pa->momentum, nb = pb->momentum;

      //----- parameters -----
      double Q2 = 2.0*(na*nb), z = vars.z, y = vars.y;
      double xa = 1.0/z-1.0-y, xb = z*y;
      double mT = std::sqrt(xa*xb);

      //----- Now we have every parameters we can define the momenta -----
      lorentzvector<double> kT(mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0, 0.0);
      
      //------ azimuthal angle is measured from the color connected parton -----      
      if(vars.spectator != vars.emitter) kT.rotateZ(p[vars.spectator].momentum.phi());
      
      //----- the daughter momenta -----
      pa->flavor = vars.fl;
      pa->momentum = na/z;
      
      pn->flavor = vars.fm;
      pn->momentum = xa*na + xb*nb + std::sqrt(Q2)*kT;
      
      //----- boost parameters -----
      double omega = 1.0/z-xa;
      
      //----- do the Lorentz boost -----
      __details::boost_ini(na, nb, omega, -kT, p);
 
      /*   returns the Lorentz boost   */
      return [=](const lorentzvector<double>& qnull) -> lorentzvector<double> {
        double nn = na*nb;
        double rs = std::sqrt(2.0*nn);
        
        //----- boosting the qnull momentum -----
        double alpha = qnull*nb/nn, beta = qnull*na/nn;
        lorentzvector<double> pT = qnull - alpha*na - beta*nb;
        
        beta -= kT.mag2()*alpha - 2.0*(kT*pT)/rs;
        pT -= alpha*rs*kT;
        
        return (alpha*omega)*na + (beta/omega)*nb + pT;
      };
    }

  
  
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping(event<decay>& p, const event<decay>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < 1)
        throw "shower_mapping(event<decay>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*  do the mapping  */
      auto boost = __details::momentum_mapping_fin(p, q, vars);
      
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
      
      return boost;
    }
    
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping(event<epa>& p, const event<epa>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < 1)
        throw "shower_mapping(event<epa>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*  do the mapping  */
      auto boost = __details::momentum_mapping_fin(p, q, vars);
      
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
      
      return boost;
    }
    
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping(event<dis>& p, const event<dis>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < 0)
        throw "shower_mapping(event<dis>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*   return value   */
      std::function<lorentzvector<double>(const lorentzvector<double>&)> boost;
      
      
      /*   do the mapping  */
      if(0 == vars.emitter)
        boost = __details::momentum_mapping_ini(p, q[-1].momentum, q[hadron(0)].momentum, q, vars);
      else boost = __details::momentum_mapping_fin(p, q, vars);
      
      //----- copying the hadron -----
      p[hadron(0)] = q[hadron(0)];
      
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
      
      //------ gluon splits to antiquark-quark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -3 || vars.sigmaP == -3))
        p.color_suppression_index += 2U;
      
      return boost;
    }
    
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping(event<hhc>& p, const event<hhc>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < -1)
        throw "shower_mapping(event<hhc>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*   return value   */
      std::function<lorentzvector<double>(const lorentzvector<double>&)> boost;
      
      /*   do the mapping   */
      if(vars.emitter < 1)
        boost = __details::momentum_mapping_ini(p, q[hadron(-1)].momentum, q[hadron(0)].momentum, q, vars);
      else boost = __details::momentum_mapping_fin(p, q, vars);
      
      //----- copying the hadron -----
      p[hadron(-1)] = q[hadron(-1)];
      p[hadron(0)] = q[hadron(0)];
      
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
      
      //------ gluon splits to antiquark-quark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -3 || vars.sigmaP == -3)) 
        p.color_suppression_index += 2U;
      
      return boost;
    }
    
  
    
    
    
    
    /**********************/
    
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping_no_color(event<decay>& p, const event<decay>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < 1)
        throw "shower_mapping(event<decay>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*  do the mapping  */
      return __details::momentum_mapping_fin(p, q, vars);
    }
    
    
    template<>
    void shower_color_mapping(event<decay>& p, const event<decay>& q, const splitting_variables<shower<lo>>& vars)
    {
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
    }
    
    

    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping_no_color(event<epa>& p, const event<epa>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < 1)
        throw "shower_mapping(event<epa>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*  do the mapping  */
      return __details::momentum_mapping_fin(p, q, vars);
    }
    
    
    template<>
    void shower_color_mapping(event<epa>& p, const event<epa>& q, const splitting_variables<shower<lo>>& vars)
    {
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
    }
    
    
    
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping_no_color(event<dis>& p, const event<dis>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < 0)
        throw "shower_mapping(event<dis>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*   copying the hadron    */
      p[hadron(0)] = q[hadron(0)];
      
      /*   do the mapping  */
      if(0 == vars.emitter)
        return __details::momentum_mapping_ini(p, q[-1].momentum, q[hadron(0)].momentum, q, vars);
      
      return __details::momentum_mapping_fin(p, q, vars);
    }

    
    template<>
    void shower_color_mapping(event<dis>& p, const event<dis>& q, const splitting_variables<shower<lo>>& vars)
    {
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
      
      //------ gluon splits to antiquark-quark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -3 || vars.sigmaP == -3))
        p.color_suppression_index += 2U;
    }
    
    
  
    
    template<> std::function<lorentzvector<double>(const lorentzvector<double>&)>
    shower_mapping_no_color(event<hhc>& p, const event<hhc>& q, const splitting_variables<shower<lo>>& vars)
    {
      if(vars.emitter < -1)
        throw "shower_mapping(event<hhc>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
      
      /*   copying the hadron    */
      p[hadron(-1)] = q[hadron(-1)];
      p[hadron(0)] = q[hadron(0)];
      
      /*   do the mapping   */
      if(vars.emitter < 1)
        return __details::momentum_mapping_ini(p, q[hadron(-1)].momentum, q[hadron(0)].momentum, q, vars);
      
      
      return __details::momentum_mapping_fin(p, q, vars);
    }
    
    
    template<>
    void shower_color_mapping(event<hhc>& p, const event<hhc>& q, const splitting_variables<shower<lo>>& vars)
    {
      //----- assigning the new color state ----
      __details::color_a_dagger(vars.sigma, vars.emitter, p);
      __details::color_a(vars.sigmaP, vars.emitter, p);
      
      //----- issuing the health report for the new state ----
      p.make_health_report();
      
      //----- healthy gluon, crossed configuration -----
      if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
        p.color_suppression_index += 2U;
      
      //------ gluon splits to quark-antiquark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2))
        p.color_suppression_index += 2U;
      
      //------ gluon splits to antiquark-quark ----
      if(q[vars.emitter].health == 'h' && (vars.sigma == -3 || vars.sigmaP == -3))
        p.color_suppression_index += 2U;
    }
    

  }    /*  namespace __details   */
}   /*   namespace duct   */

